from ..core import WesternCalendar
from ..registry_tools import iso_register


@iso_register('CZ')
class CzechRepublic(WesternCalendar):
    'Czech Republic'

    # Civil holidays
    include_labour_day = True

    # Christian holidays
    include_easter_monday = True
    include_good_friday = True

    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (1, 1, "Restoration Day of the Independent Czech State"),
        (5, 8, "Liberation Day"),
        (7, 5, "Saints Cyril and Methodius Day"),
        (7, 6, "Jan Hus Day"),
        (9, 28, "St. Wenceslas Day (Czech Statehood Day)"),
        (10, 28, "Independent Czechoslovak State Day"),
        (11, 17, "Struggle for Freedom and Democracy Day"),
        (12, 24, "Christmas Eve"),
        (12, 26, "St. Stephen's Day (The Second Christmas Day)"),
    )

    def get_variable_days(self, year):
        # As of 2016, Good Friday became a holiday
        self.include_good_friday = (year >= 2016)
        return super().get_variable_days(year)



import os
import tempfile
import warnings
from datetime import date
from pathlib import Path
from unittest import TestCase
from platform import system

from freezegun import freeze_time

from ..core import Calendar
from .. import __version__


class CoreCalendarTest(TestCase):
    cal_class = Calendar
    kwargs = {}

    def setUp(self):
        super().setUp()
        self.year = date.today().year
        self.cal = self.cal_class(**self.kwargs)


class GenericCalendarTest(CoreCalendarTest):
    test_include_january_1st = True

    def setUp(self):
        super().setUp()
        warnings.simplefilter('ignore')

    def test_weekend_days(self):
        class_name = self.cal_class.__name__
        if class_name in ['Calendar']:
            return
        try:
            self.cal.get_weekend_days()
        except NotImplementedError:
            assert False, (self.cal, class_name)

    def test_january_1st(self):
        class_name = self.cal_class.__name__
        if class_name in ['Calendar']:
            return
        holidays = self.cal.holidays_set(self.year)
        if self.test_include_january_1st:
            self.assertIn(date(self.year, 1, 1), holidays)
        else:
            self.assertNotIn(date(self.year, 1, 1), holidays)

    def test_ical_export(self):
        """Check that an iCal file can be created according to iCal spec."""
        class_name = self.cal_class.__name__
        if class_name in ['Calendar']:
            return

        holidays = self.cal.holidays(2019) + self.cal.holidays(2020)

        temp_dir = Path(tempfile.gettempdir()) / "failed_ical_tests"
        temp_dir.mkdir(parents=True, exist_ok=True)
        test_file_fd, test_file_name = tempfile.mkstemp(
            prefix=f"{self.cal_class.__name__}_",
            suffix=".ics",
            dir=temp_dir,
        )
        test_path = Path(test_file_name)

        self.cal.export_to_ical(
            period=[2019, 2020],
            target_path=test_path,
        )
        # A standard iCal extension should have been added automatically:
        with test_path.open() as ics_file:
            # check header
            assert ics_file.readline() == 'BEGIN:VCALENDAR\n'
            assert ics_file.readline() == 'VERSION:2.0\n'
            prod_id_line = ics_file.readline()
            assert prod_id_line == (
                f'PRODID:-//workalendar//ical {__version__}//EN\n')
            # check new year
            assert ics_file.readline() == 'BEGIN:VEVENT\n'
            first_event_name = holidays[0][1]
            assert ics_file.readline() == f'SUMMARY:{first_event_name}\n'
            if self.test_include_january_1st:
                assert ics_file.readline() == 'DTSTART;VALUE=DATE:20190101\n'
                assert ics_file.readline().startswith(
                    'DTSTAMP;VALUE=DATE-TIME:')
                first_uid_line = ics_file.readline()
                uid_lines = [first_uid_line]
                assert first_uid_line.startswith('UID:2019')
                assert ics_file.readline() == 'END:VEVENT\n'
            else:
                uid_lines = []
            remaining_lines = ics_file.readlines()
            uid_lines += [line for line in remaining_lines
                          if line.startswith('UID:')]
            # check number of entries
            assert len(uid_lines) == len(holidays)
            # check that UIDs are unique within the calendar
            assert len(uid_lines) == len(set(uid_lines))
            # check that final year is included
            assert remaining_lines[-3].startswith('UID:2020')
            # check last few lines of file
            assert remaining_lines[-2] == 'END:VEVENT\n'
            assert remaining_lines[-1] == 'END:VCALENDAR\n'

        # Freeze time to make sure they both have the same timestamp
        with freeze_time():
            # Regenerate the file
            self.cal.export_to_ical(
                period=[2019, 2020],
                target_path=test_path,
            )
            # Not providing any target => returns the value as a string
            var_contents = self.cal.export_to_ical(period=[2019, 2020])

        file_contents = test_path.read_text()

        assert file_contents == var_contents

        # If platform is Windows close the temp file descriptor
        if system() == 'Windows':  # pragma: no cover
            os.close(test_file_fd)
        # Remove the .ics file if this test passes
        test_path.unlink()



from .core import UnitedStates
from ..registry_tools import iso_register


@iso_register('US-MO')
class Missouri(UnitedStates):
    """Missouri"""
    FIXED_HOLIDAYS = UnitedStates.FIXED_HOLIDAYS + (
        (5, 8, "Truman Day"),
    )
    include_lincoln_birthday = True



from ..registry_tools import iso_register
from .core import UnitedStates


@iso_register('US-NJ')
class NewJersey(UnitedStates):
    """New Jersey"""
    include_good_friday = True
    include_election_day_every_year = True



from ..core import IslamicCalendar
from ..registry_tools import iso_register


@iso_register('DZ')
class Algeria(IslamicCalendar):
    "Algeria"
    # Civil holidays
    include_labour_day = True
    # Islamic holidays
    include_prophet_birthday = True
    include_eid_al_fitr = True
    include_day_of_sacrifice = True
    include_islamic_new_year = True

    FIXED_HOLIDAYS = IslamicCalendar.FIXED_HOLIDAYS + (
            (7, 5, "Independence Day"),
            (11, 1, "Anniversary of the revolution"),
        )

    ISLAMIC_HOLIDAYS = IslamicCalendar.ISLAMIC_HOLIDAYS + (
        (1, 10, "Ashura"),
    )



import sys

if sys.version_info[:2] < (3, 8):  # coverage: exclude
    import importlib_metadata
else:  # coverage: exclude
    import importlib.metadata as importlib_metadata

try:
    #: Module version, as defined in PEP-0396.
    __version__ = importlib_metadata.version(__name__)
except importlib_metadata.PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



from .core import UnitedStates
from ..registry_tools import iso_register


@iso_register('US-LA')
class Louisiana(UnitedStates):
    """Louisiana"""
    include_good_friday = True
    include_election_day_even = True
    include_columbus_day = False
    include_fat_tuesday = True



from datetime import date

from . import GenericCalendarTest
from ..europe import (
    Spain, Andalusia, Aragon, Catalonia, CastileAndLeon, CastillaLaMancha,
    CanaryIslands, Extremadura, Galicia, BalearicIslands, LaRioja,
    CommunityofMadrid, Murcia, Navarre, Asturias, BasqueCountry, Cantabria,
    ValencianCommunity
)


class SpainTest(GenericCalendarTest):
    cal_class = Spain

    def test_common_year_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 1, 1), holidays)    # New Year
        self.assertIn(date(2015, 1, 6), holidays)    # Epiphany
        self.assertIn(date(2015, 4, 3), holidays)    # Good Friday
        self.assertIn(date(2015, 5, 1), holidays)    # Labour Day
        self.assertIn(date(2015, 8, 15), holidays)   # Assumption
        self.assertIn(date(2015, 10, 12), holidays)  # Nation Day
        self.assertIn(date(2015, 11, 1), holidays)   # All Saints
        self.assertIn(date(2015, 12, 6), holidays)   # Dia de la Constitución
        self.assertIn(date(2015, 12, 8), holidays)   # Immaculate conception
        self.assertIn(date(2015, 12, 25), holidays)  # Christmas

    def test_common_year_2016(self):
        holidays = self.cal.holidays_set(2016)
        self.assertIn(date(2016, 1, 1), holidays)    # New Year
        self.assertIn(date(2016, 1, 6), holidays)    # Epiphany
        self.assertIn(date(2016, 3, 25), holidays)   # Good Friday
        self.assertIn(date(2016, 5, 1), holidays)    # Labour Day
        self.assertIn(date(2016, 8, 15), holidays)   # Assumption
        self.assertIn(date(2016, 10, 12), holidays)  # Nation Day
        self.assertIn(date(2016, 11, 1), holidays)   # All Saints
        self.assertIn(date(2016, 12, 6), holidays)   # Dia de la Constitución
        self.assertIn(date(2016, 12, 8), holidays)   # Immaculate conception
        self.assertIn(date(2016, 12, 25), holidays)  # Christmas

    def test_labour_day_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        self.assertEqual(
            holidays[date(2020, 5, 1)], "Día del trabajador")


class AndalusiaTest(SpainTest):
    cal_class = Andalusia

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 2, 28), holidays)  # Andalusian National Day
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        self.assertEqual(len(holidays), 12)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 2, 28), holidays)  # Andalusian National Day
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        self.assertEqual(len(holidays), 12)


class AragonTest(SpainTest):
    cal_class = Aragon

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2019, 4, 23), holidays)  # Regional Day
        self.assertIn(date(2019, 12, 20), holidays)  # Aragon Ombudsman Day
        self.assertEqual(len(holidays), 13)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2020, 4, 23), holidays)  # Regional Day
        self.assertIn(date(2020, 12, 20), holidays)  # Aragon Ombudsman Day
        self.assertEqual(len(holidays), 13)


class CastileAndLeonTest(SpainTest):
    cal_class = CastileAndLeon

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2019, 4, 23), holidays)  # Regional Day
        self.assertEqual(len(holidays), 12)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2020, 4, 23), holidays)  # Regional Day
        self.assertEqual(len(holidays), 12)


class CastillaLaManchaTest(SpainTest):
    cal_class = CastillaLaMancha

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2019, 5, 31), holidays)  # Regional Day
        self.assertEqual(len(holidays), 12)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2020, 5, 31), holidays)  # Regional Day
        self.assertEqual(len(holidays), 12)


class CanaryIslandsTest(SpainTest):
    cal_class = CanaryIslands

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2019, 5, 30), holidays)  # Regional Day
        self.assertEqual(len(holidays), 12)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2020, 5, 30), holidays)  # Regional Day
        self.assertEqual(len(holidays), 12)


class CataloniaTest(SpainTest):
    cal_class = Catalonia

    def test_region_year_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 4, 6), holidays)  # Easter Monday
        # Diada nacional de Catalunya
        self.assertIn(date(2015, 9, 11), holidays)
        # La revetlla de Sant Joan, Nit de Sant Joan
        self.assertIn(date(2015, 6, 24), holidays)
        self.assertIn(date(2015, 12, 26), holidays)  # Sant Esteve
        self.assertEqual(len(holidays), 14)

    def test_region_year_2016(self):
        holidays = self.cal.holidays_set(2016)
        self.assertIn(date(2016, 3, 28), holidays)  # Easter Monday
        # Diada nacional de Catalunya
        self.assertIn(date(2016, 9, 11), holidays)
        # La revetlla de Sant Joan, Nit de Sant Joan
        self.assertIn(date(2016, 6, 24), holidays)
        self.assertIn(date(2016, 12, 26), holidays)  # Sant Esteve
        self.assertEqual(len(holidays), 14)

    def test_boxing_day_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        self.assertEqual(holidays[date(2020, 12, 26)], "Sant Esteve")


class ExtremaduraTest(SpainTest):
    cal_class = Extremadura

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2019, 9, 8), holidays)  # Regional Day
        self.assertEqual(len(holidays), 12)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2020, 9, 8), holidays)  # Regional Day
        self.assertEqual(len(holidays), 12)


class GaliciaTest(SpainTest):
    cal_class = Galicia

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2019, 5, 17), holidays)  # Galician Literature Day
        self.assertIn(date(2019, 7, 25), holidays)  # Regional Day
        self.assertEqual(len(holidays), 13)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2020, 5, 17), holidays)  # Galician Literature Day
        self.assertIn(date(2020, 7, 25), holidays)  # Regional Day
        self.assertEqual(len(holidays), 13)


class BalearicIslandsTest(SpainTest):
    cal_class = BalearicIslands

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 3, 1), holidays)  # Balearic National Day
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2019, 4, 22), holidays)  # Easter Monday
        self.assertEqual(len(holidays), 13)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 3, 1), holidays)  # Balearic National Day
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2020, 4, 13), holidays)  # Easter Monday
        self.assertEqual(len(holidays), 13)


class LaRiojaTest(SpainTest):
    cal_class = LaRioja

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2019, 6, 9), holidays)  # Dia de La Rioja
        self.assertEqual(len(holidays), 12)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2020, 6, 9), holidays)  # Dia de La Rioja
        self.assertEqual(len(holidays), 12)


class CommunityofMadridTest(SpainTest):
    cal_class = CommunityofMadrid

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        # Fiesta de la Comunidad de Madrid
        self.assertIn(date(2019, 5, 2), holidays)
        self.assertEqual(len(holidays), 12)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        # Fiesta de la Comunidad de Madrid
        self.assertIn(date(2020, 5, 2), holidays)
        self.assertEqual(len(holidays), 12)


class MurciaTest(SpainTest):
    cal_class = Murcia

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 3, 19), holidays)  # San José
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2019, 6, 9), holidays)  # Día de la Región de Murcia
        self.assertEqual(len(holidays), 13)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 3, 19), holidays)  # San José
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2020, 6, 9), holidays)  # Día de la Región de Murcia
        self.assertEqual(len(holidays), 13)


class NavarreTest(SpainTest):
    cal_class = Navarre

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2019, 4, 22), holidays)  # Easter Monday
        self.assertEqual(len(holidays), 12)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2020, 4, 13), holidays)  # Easter Monday
        self.assertEqual(len(holidays), 12)


class AsturiasTest(SpainTest):
    cal_class = Asturias

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2019, 9, 8), holidays)  # Día de Asturias
        self.assertEqual(len(holidays), 12)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2020, 9, 8), holidays)  # Día de Asturias
        self.assertEqual(len(holidays), 12)


class BasqueCountryTest(SpainTest):
    cal_class = BasqueCountry

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2019, 4, 22), holidays)  # Easter Monday
        self.assertIn(date(2019, 10, 25), holidays)  # Euskadi Eguna
        self.assertEqual(len(holidays), 13)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        self.assertIn(date(2020, 4, 13), holidays)  # Easter Monday
        self.assertIn(date(2020, 10, 25), holidays)  # Euskadi Eguna
        self.assertEqual(len(holidays), 13)


class CantabriaTest(SpainTest):
    cal_class = Cantabria

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy/Holy Thursday
        # Día de Cantabria o Día de La Montaña
        self.assertIn(date(2019, 9, 15), holidays)
        self.assertEqual(len(holidays), 12)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 4, 9), holidays)  # Maundy/Holy Thursday
        # Día de Cantabria o Día de La Montaña
        self.assertIn(date(2020, 9, 15), holidays)
        self.assertEqual(len(holidays), 12)


class ValencianCommunityTest(SpainTest):
    cal_class = ValencianCommunity

    def test_region_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 3, 19), holidays)  # San José
        self.assertIn(date(2019, 4, 22), holidays)  # Easter Monday
        # Dia de la Comunitat Valenciana
        self.assertIn(date(2019, 10, 9), holidays)
        self.assertEqual(len(holidays), 13)

    def test_region_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 3, 19), holidays)  # San José
        self.assertIn(date(2020, 4, 13), holidays)  # Easter Monday
        # Dia de la Comunitat Valenciana
        self.assertIn(date(2020, 10, 9), holidays)
        self.assertEqual(len(holidays), 13)



from datetime import timedelta
from ..core import IslamicCalendar, SAT, SUN
from ..registry_tools import iso_register


@iso_register('TR')
class Turkey(IslamicCalendar):
    'Turkey'
    shift_new_years_day = True
    # Even though they're using an islamic calendar, the work week is MON->FRI
    WEEKEND_DAYS = (SAT, SUN)

    # Islamic Holidays
    include_eid_al_fitr = True
    length_eid_al_fitr = 3
    include_eid_al_adha = True
    length_eid_al_adha = 4

    # Civil holidays
    include_labour_day = True
    labour_day_label = "Labor and Solidarity Day"
    FIXED_HOLIDAYS = IslamicCalendar.FIXED_HOLIDAYS + (
        (4, 23, "National Sovereignty and Children's Day"),
        (5, 19, "Commemoration of Atatürk, Youth and Sports Day"),
        (7, 15, "Democracy and National Unity Day"),
        (8, 30, "Victory Day"),
        (10, 29, "Republic Day"),
    )

    def get_delta_islamic_holidays(self, year):
        """
        Turkey Islamic holidays are shifted by one day every year.
        """
        return timedelta(days=-1)



from .core import UnitedStates
from ..registry_tools import iso_register


@iso_register('US-CO')
class Colorado(UnitedStates):
    """Colorado"""
    # Colorado has only federal state holidays.
    # NOTE: Cesar Chavez Day is an optional holiday



from datetime import date, timedelta

from ..core import WesternCalendar, MON, SAT, SUN
from ..registry_tools import iso_register


@iso_register("NZ")
class NewZealand(WesternCalendar):
    "New Zealand"
    include_good_friday = True
    include_easter_monday = True
    include_boxing_day = True

    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (1, 2, "Day after New Year's Day"),
        (2, 6, "Waitangi Day"),
        (4, 25, "ANZAC Day")
    )

    def get_queens_birthday(self, year):
        return (
            NewZealand.get_nth_weekday_in_month(year, 6, MON, 1),
            "Queen's Birthday"
        )

    def get_labour_day(self, year):
        return (
            NewZealand.get_nth_weekday_in_month(year, 10, MON, 4),
            "Labour Day"
        )

    def get_variable_days(self, year):
        # usual variable days
        days = super().get_variable_days(year)
        days.append(self.get_queens_birthday(year))
        days.append(self.get_labour_day(year))

        waitangi_day = date(year, 2, 6)
        if waitangi_day.weekday() in self.get_weekend_days():
            days.append((
                self.find_following_working_day(waitangi_day),
                "Waitangi Day Shift")
            )

        anzac_day = date(year, 4, 25)
        if anzac_day.weekday() in self.get_weekend_days():
            days.append((
                self.find_following_working_day(anzac_day),
                "ANZAC Day Shift")
            )

        christmas = date(year, 12, 25)
        boxing_day = date(year, 12, 26)
        if christmas.weekday() == SAT:
            shift = self.find_following_working_day(christmas)
            days.append((shift, "Christmas Shift"))
        elif christmas.weekday() == SUN:
            shift = self.find_following_working_day(christmas)
            days.append((shift + timedelta(days=1), "Christmas Shift"))

        if boxing_day.weekday() == SAT:
            shift = self.find_following_working_day(boxing_day)
            days.append((shift, "Boxing Day Shift"))
        elif boxing_day.weekday() == SUN:
            shift = self.find_following_working_day(boxing_day)
            days.append((shift + timedelta(days=1), "Boxing Day Shift"))

        new_year = date(year, 1, 1)
        day_after_new_year = date(year, 1, 2)
        if new_year.weekday() == SAT:
            shift = self.find_following_working_day(new_year)
            days.append((shift, "New Year Shift"))
        elif new_year.weekday() == SUN:
            shift = self.find_following_working_day(new_year)
            days.append((shift + timedelta(days=1), "New Year Shift"))

        if day_after_new_year.weekday() == SAT:
            shift = self.find_following_working_day(day_after_new_year)
            days.append((shift, "Day after New Year's Day Shift"))
        elif day_after_new_year.weekday() == SUN:
            shift = self.find_following_working_day(day_after_new_year)
            days.append((shift + timedelta(days=1),
                         "Day after New Year's Day Shift"))

        return days



from ..core import WesternCalendar, SUN
from ..registry_tools import iso_register


@iso_register('LT')
class Lithuania(WesternCalendar):
    'Lithuania'

    # Civil holidays
    include_labour_day = True
    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (2, 16, "Restoration of the State Day"),
        (3, 11, "Restoration of Independence Day"),
        (6, 24, "St. John's Day"),
        (7, 6, "Anniversary of the Coronation of King Mindaugas"),
    )

    # Christian holidays
    include_easter_sunday = True
    include_easter_monday = True
    include_assumption = True
    include_all_saints = True
    include_christmas_eve = True
    include_christmas = True
    include_boxing_day = True
    boxing_day_label = "Second day of Christmas"

    def get_mothers_day(self, year):
        return (
            Lithuania.get_nth_weekday_in_month(year, 5, SUN, 1),
            "Mother's day"
        )

    def get_fathers_day(self, year):
        return (
            Lithuania.get_nth_weekday_in_month(year, 6, SUN, 1),
            "Father's day"
        )

    def get_variable_days(self, year):
        # All Souls day was introduced as of 2020
        # https://en.wikipedia.org/wiki/Public_holidays_in_Lithuania
        self.include_all_souls = year >= 2020
        days = super().get_variable_days(year)
        days.append(self.get_mothers_day(year))
        days.append(self.get_fathers_day(year))
        return days



from datetime import date
from ..core import WesternCalendar, FRI, SAT
from ..registry_tools import iso_register


@iso_register('SE')
class Sweden(WesternCalendar):
    'Sweden'

    # Christian holidays
    include_epiphany = True
    include_good_friday = True
    include_easter_sunday = True
    include_easter_monday = True
    include_ascension = True
    include_whit_sunday = True
    whit_sunday_label = "Pentecost"
    # Christmas Eve is not a holiday but not a work day either
    include_christmas_eve = True
    include_boxing_day = True
    boxing_day_label = "Second Day of Christmas"

    # Civil holidays
    include_labour_day = True
    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (6, 6, "National Day"),
        # New Year's Eve is not a holiday but not a work day either
        (12, 31, "New Year's Eve")
    )

    # Midsummer Eve is not a holiday but not a work day either
    def get_midsummer_eve(self, year):
        date_eve = Sweden.get_nth_weekday_in_month(
            year, 6, FRI, start=date(year, 6, 19))
        return date_eve

    def get_midsummer_day(self, year):
        date_eve = Sweden.get_nth_weekday_in_month(
            year, 6, SAT, start=date(year, 6, 20))
        return date_eve

    def get_variable_all_saints(self, year):
        all_saints = date(year, 10, 31)
        if all_saints.weekday() != SAT:
            all_saints = Sweden.get_nth_weekday_in_month(
                year, 11, SAT)
        return all_saints

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.append((self.get_midsummer_day(year), "Midsummer's Day"))
        days.append((self.get_midsummer_eve(year), "Midsummer's Eve"))
        days.append((self.get_variable_all_saints(year), "All Saints"))
        return days



from datetime import date
from ..core import WesternCalendar
from ..registry_tools import iso_register


@iso_register('LV')
class Latvia(WesternCalendar):
    'Latvia'

    # Civil holidays
    include_labour_day = True
    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (6, 23, "Midsummer Day"),
        (6, 24, "St. John's Day"),
        (11, 18, "Proclamation Day"),
        (12, 31, "New Years Eve"),
    )

    # Christian holidays
    include_good_friday = True
    include_easter_sunday = True
    include_easter_monday = True
    include_christmas_eve = True
    include_christmas = True
    include_boxing_day = True

    def get_independence_days(self, year):
        """returns a possibly empty list of (date, holiday_name) tuples"""
        days = []
        if year > 2004:
            actual_date = date(year, 5, 4)
            days = [(actual_date, "Restoration of Independence Day")]
            if actual_date.weekday() in self.get_weekend_days():
                days += [(self.find_following_working_day(actual_date),
                          "Restoration of Independence Observed")]
        return days

    def get_republic_days(self, year):
        """returns a possibly empty list of (date, holiday_name) tuples"""
        days = []
        if year > 1918:
            actual_date = date(year, 11, 18)
            days = [(actual_date, "Proclamation of Republic Day")]
            if actual_date.weekday() in self.get_weekend_days():
                days += [(self.find_following_working_day(actual_date),
                          "Proclamation of Republic Observed")]
        return days

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.extend(self.get_independence_days(year))
        days.extend(self.get_republic_days(year))
        return days



from datetime import date, timedelta
import warnings
from pyluach.dates import GregorianDate

from .core import UnitedStates
from ..registry_tools import iso_register


class HebrewHolidays:

    hebrew_calendars = {}

    @classmethod
    def get_hebrew_calendar(cls, gregorian_year):
        """
        Build and cache the Hebrew calendar for the given Gregorian Year.
        """
        if gregorian_year not in cls.hebrew_calendars:
            # Build the hebrew calendar for year
            days = []
            current_date = date(gregorian_year, 1, 1)

            while current_date.year == gregorian_year:
                hebrew_date = GregorianDate(
                    year=current_date.year,
                    month=current_date.month,
                    day=current_date.day,
                ).to_heb()
                days.append(
                    (hebrew_date, current_date)
                )
                current_date += timedelta(days=1)
            # Store it in the class property
            cls.hebrew_calendars[gregorian_year] = days

        # Return the hebrew calendar
        return cls.hebrew_calendars[gregorian_year]

    @classmethod
    def search_hebrew_calendar(cls, gregorian_year, hebrew_month, hebrew_day):
        """
        Search for a specific Hebrew month and day in the Hebrew calendar.
        """
        calendar = cls.get_hebrew_calendar(gregorian_year)
        search = filter(lambda item: item[0].month == hebrew_month, calendar)
        search = filter(lambda item: item[0].day == hebrew_day, search)
        for item in search:
            return item[1]

    @classmethod
    def get_rosh_hashanah(cls, year):
        """
        Return the gregorian date of the first day of Rosh Hashanah
        """
        return cls.search_hebrew_calendar(year, 7, 1)

    @classmethod
    def get_yom_kippur(cls, year):
        """
        Return the gregorian date of Yom Kippur.
        """
        return cls.search_hebrew_calendar(year, 7, 10)


@iso_register('US-FL')
class Florida(UnitedStates):
    """Florida"""
    include_thanksgiving_friday = True
    thanksgiving_friday_label = "Friday after Thanksgiving"
    include_columbus_day = False
    include_federal_presidents_day = False


class FloridaLegal(Florida):
    """Florida Legal Holidays"""
    FIXED_HOLIDAYS = Florida.FIXED_HOLIDAYS + (
        (2, 15, 'Susan B. Anthony Day'),
        (4, 2, 'Pascua Florida Day'),
        (6, 14, 'Flag Day'),
    )
    include_fat_tuesday = True
    include_lincoln_birthday = True
    include_federal_presidents_day = True
    include_good_friday = True
    include_confederation_day = True
    include_jefferson_davis_birthday = True
    include_columbus_day = True
    columbus_day_label = "Columbus Day and Farmers' Day"
    include_election_day_every_year = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn(
            "Florida's laws separate the definitions between paid versus legal"
            " holidays. Be warned that Florida Legal specific Holidays are not"
            " paid holidays."
        )

    def get_confederate_day(self, year):
        """
        Confederation memorial day is on the April 26th for Florida Legal.
        """
        return date(year, 4, 26), "Confederate Memorial Day"

    def get_jefferson_davis_birthday(self, year):
        """
        Jefferson Davis Birthday appears to be a fixed holiday (June 3rd)
        """
        return date(year, 6, 3), "Jefferson Davis Birthday"


class FloridaCircuitCourts(HebrewHolidays, Florida):
    """Florida Circuits Courts"""
    include_federal_presidents_day = True
    include_good_friday = True

    def get_variable_days(self, year):
        days = super().get_variable_days(year)

        days.append((
            self.get_rosh_hashanah(year),
            "Rosh Hashanah"
        ))

        days.append((
            self.get_yom_kippur(year),
            "Yom Kippur"
        ))

        return days


class FloridaMiamiDade(Florida):
    """Miami-Dade, Florida"""
    include_federal_presidents_day = True
    include_columbus_day = True



from datetime import date

from ..core import MON, TUE, WED, THU, FRI, SAT, SUN
from ..registry_tools import iso_register
from .core import UnitedStates


@iso_register('US-NC')
class NorthCarolina(UnitedStates):
    """North Carolina"""
    include_good_friday = True
    include_christmas_eve = True
    include_thanksgiving_friday = True
    include_boxing_day = True
    include_federal_presidents_day = False
    include_columbus_day = False

    def get_christmas_shifts(self, year):
        """
        Return Specific Christmas days extra shifts.
        There must be 3 holidays in a row: Christmas Eve, Christmas Day and
        Boxing Day. If one or the other falls on SUN/SAT, extra days must be
        added.
        """
        xmas = date(year, 12, 25)
        if xmas.weekday() in (TUE, WED, THU):
            # No shift, move along
            return []
        if xmas.weekday() == FRI:
            return [
                (date(year, 12, 28), "Boxing day shift"),
            ]
        elif xmas.weekday() == SAT:
            return [
                (date(year, 12, 23), "Christmas Eve shift"),
                (date(year, 12, 27), "Boxing Day shift"),
            ]
        elif xmas.weekday() == SUN:
            return [
                (date(year, 12, 23), "Christmas Eve shift"),
                (date(year, 12, 27), "Boxing Day shift"),
            ]
        elif xmas.weekday() == MON:
            return [
                (date(year, 12, 27), "Christmas Eve shift"),
            ]

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.extend(self.get_christmas_shifts(year))
        return days



"""
Working day tools
"""
from copy import copy
import warnings
from calendar import monthrange
from datetime import date, timedelta, datetime
from pathlib import Path
import sys

import convertdate
from dateutil import easter
from lunardate import LunarDate

from .exceptions import (
    UnsupportedDateType, CalendarError,
    ICalExportRangeError, ICalExportTargetPathError
)
from . import __version__

MON, TUE, WED, THU, FRI, SAT, SUN = range(7)
ISO_MON, ISO_TUE, ISO_WED, ISO_THU, ISO_FRI, ISO_SAT, ISO_SUN = range(1, 8)


class classproperty:

    def __init__(self, getter):
        self.getter = getter
        self.__doc__ = getter.__doc__

    def __get__(self, instance, owner):
        return self.getter(owner)


def cleaned_date(day, keep_datetime=False):
    """
    Return a "clean" date type.

    * keep a `date` unchanged
    * convert a datetime into a date,
    * convert any "duck date" type into a date using its `date()` method.
    """
    if not isinstance(day, (date, datetime)):
        raise UnsupportedDateType(
            f"`{day}` is of unsupported type ({type(day)})"
        )
    if not keep_datetime:
        if hasattr(day, 'date') and callable(day.date):
            day = day.date()
    return day


def daterange(start, end):
    """
    Yield days from ``start`` to ``end`` including both of them.

    If start and end are in opposite order, they'll be swapped silently.
    """
    # Swap if necessary
    if start > end:
        end, start = start, end
    day = start
    while day <= end:
        yield day
        day += timedelta(days=1)


class ChristianMixin:
    EASTER_METHOD = None  # to be assigned in the inherited mixin
    include_epiphany = False
    include_clean_monday = False
    include_annunciation = False
    include_fat_tuesday = False
    # Fat tuesday forced to `None` to make sure this value is always set
    # We've seen that there was a wide variety of labels.
    fat_tuesday_label = None
    include_ash_wednesday = False
    ash_wednesday_label = "Ash Wednesday"
    include_palm_sunday = False
    include_holy_thursday = False
    holy_thursday_label = "Holy Thursday"
    include_good_friday = False
    good_friday_label = "Good Friday"
    include_easter_monday = False
    include_easter_saturday = False
    easter_saturday_label = "Easter Saturday"
    include_easter_sunday = False
    include_all_saints = False
    include_immaculate_conception = False
    immaculate_conception_label = "Immaculate Conception"
    include_christmas = True
    christmas_day_label = "Christmas Day"
    include_christmas_eve = False
    include_ascension = False
    include_assumption = False
    include_whit_sunday = False
    whit_sunday_label = 'Whit Sunday'
    include_whit_monday = False
    whit_monday_label = 'Whit Monday'
    include_corpus_christi = False
    include_boxing_day = False
    boxing_day_label = "Boxing Day"
    include_all_souls = False

    def get_fat_tuesday(self, year):
        if not self.fat_tuesday_label:
            raise CalendarError(
                "Improperly configured: please provide a "
                "`fat_tuesday_label` value")
        sunday = self.get_easter_sunday(year)
        return sunday - timedelta(days=47)

    def get_ash_wednesday(self, year):
        sunday = self.get_easter_sunday(year)
        return sunday - timedelta(days=46)

    def get_palm_sunday(self, year):
        sunday = self.get_easter_sunday(year)
        return sunday - timedelta(days=7)

    def get_holy_thursday(self, year):
        "Return the date of the last thursday before easter"
        sunday = self.get_easter_sunday(year)
        return sunday - timedelta(days=3)

    def get_good_friday(self, year):
        "Return the date of the last friday before easter"
        sunday = self.get_easter_sunday(year)
        return sunday - timedelta(days=2)

    def get_clean_monday(self, year):
        "Return the clean monday date"
        sunday = self.get_easter_sunday(year)
        return sunday - timedelta(days=48)

    def get_easter_saturday(self, year):
        "Return the Easter Saturday date"
        sunday = self.get_easter_sunday(year)
        return sunday - timedelta(days=1)

    def get_easter_sunday(self, year):
        "Return the date of the easter (sunday) -- following the easter method"
        return easter.easter(year, self.EASTER_METHOD)

    def get_easter_monday(self, year):
        "Return the date of the monday after easter"
        sunday = self.get_easter_sunday(year)
        return sunday + timedelta(days=1)

    def get_ascension_thursday(self, year):
        easter = self.get_easter_sunday(year)
        return easter + timedelta(days=39)

    def get_whit_monday(self, year):
        easter = self.get_easter_sunday(year)
        return easter + timedelta(days=50)

    def get_whit_sunday(self, year):
        easter = self.get_easter_sunday(year)
        return easter + timedelta(days=49)

    def get_corpus_christi(self, year):
        return self.get_easter_sunday(year) + timedelta(days=60)

    def shift_christmas_boxing_days(self, year):
        """ When Christmas and/or Boxing Day falls on a weekend, it is rolled
            forward to the next weekday.
        """
        christmas = date(year, 12, 25)
        boxing_day = date(year, 12, 26)
        boxing_day_label = f"{self.boxing_day_label} Shift"
        results = []
        if christmas.weekday() in self.get_weekend_days():
            shift = self.find_following_working_day(christmas)
            results.append((shift, "Christmas Shift"))
            results.append((shift + timedelta(days=1), boxing_day_label))
        elif boxing_day.weekday() in self.get_weekend_days():
            shift = self.find_following_working_day(boxing_day)
            results.append((shift, boxing_day_label))
        return results

    def get_variable_days(self, year):  # noqa
        "Return the christian holidays list according to the mixin"
        days = super().get_variable_days(year)
        if self.include_epiphany:
            days.append((date(year, 1, 6), "Epiphany"))
        if self.include_clean_monday:
            days.append((self.get_clean_monday(year), "Clean Monday"))
        if self.include_annunciation:
            days.append((date(year, 3, 25), "Annunciation"))
        if self.include_fat_tuesday:
            days.append(
                (self.get_fat_tuesday(year), self.fat_tuesday_label)
            )
        if self.include_ash_wednesday:
            days.append(
                (self.get_ash_wednesday(year), self.ash_wednesday_label)
            )
        if self.include_palm_sunday:
            days.append((self.get_palm_sunday(year), "Palm Sunday"))
        if self.include_holy_thursday:
            days.append(
                (self.get_holy_thursday(year), self.holy_thursday_label)
            )
        if self.include_good_friday:
            days.append((self.get_good_friday(year), self.good_friday_label))
        if self.include_easter_saturday:
            days.append(
                (self.get_easter_saturday(year), self.easter_saturday_label)
            )
        if self.include_easter_sunday:
            days.append((self.get_easter_sunday(year), "Easter Sunday"))
        if self.include_easter_monday:
            days.append((self.get_easter_monday(year), "Easter Monday"))
        if self.include_assumption:
            days.append((date(year, 8, 15), "Assumption of Mary to Heaven"))
        if self.include_all_saints:
            days.append((date(year, 11, 1), "All Saints Day"))
        if self.include_all_souls:
            days.append((date(year, 11, 2), "All Souls Day"))
        if self.include_immaculate_conception:
            days.append((date(year, 12, 8), self.immaculate_conception_label))
        if self.include_christmas:
            days.append((date(year, 12, 25), self.christmas_day_label))
        if self.include_christmas_eve:
            days.append((date(year, 12, 24), "Christmas Eve"))
        if self.include_boxing_day:
            days.append((date(year, 12, 26), self.boxing_day_label))
        if self.include_ascension:
            days.append((
                self.get_ascension_thursday(year), "Ascension Thursday"))
        if self.include_whit_monday:
            days.append((self.get_whit_monday(year), self.whit_monday_label))
        if self.include_whit_sunday:
            days.append((self.get_whit_sunday(year), self.whit_sunday_label))
        if self.include_corpus_christi:
            days.append((self.get_corpus_christi(year), "Corpus Christi"))
        return days


class WesternMixin(ChristianMixin):
    """
    General usage calendar for Western countries.

    (chiefly Europe and Northern America)

    """
    EASTER_METHOD = easter.EASTER_WESTERN
    WEEKEND_DAYS = (SAT, SUN)


class OrthodoxMixin(ChristianMixin):
    EASTER_METHOD = easter.EASTER_ORTHODOX
    WEEKEND_DAYS = (SAT, SUN)

    include_orthodox_christmas = True
    # This label should be de-duplicated if needed
    orthodox_christmas_day_label = "Christmas"

    def get_fixed_holidays(self, year):
        days = super().get_fixed_holidays(year)
        if self.include_orthodox_christmas:
            days.append(
                (date(year, 1, 7), self.orthodox_christmas_day_label)
            )
        return days


class LunarMixin:
    """
    Calendar ready to compute luncar calendar days
    """
    @staticmethod
    def lunar(year, month, day):
        return LunarDate(year, month, day).toSolarDate()


class ChineseNewYearMixin(LunarMixin):
    """
    Calendar including toolsets to compute the Chinese New Year holidays.
    """
    include_chinese_new_year_eve = False
    chinese_new_year_eve_label = "Chinese New Year's eve"
    # Chinese New Year will be included by default
    include_chinese_new_year = True
    chinese_new_year_label = 'Chinese New Year'
    # Some countries include the 2nd lunar day as a holiday
    include_chinese_second_day = False
    chinese_second_day_label = "Chinese New Year (2nd day)"
    include_chinese_third_day = False
    chinese_third_day_label = "Chinese New Year (3rd day)"
    shift_sunday_holidays = False
    # Some calendars roll a starting Sunday CNY to Sat
    shift_start_cny_sunday = False

    def get_chinese_new_year(self, year):
        """
        Compute Chinese New Year days. To return a list of holidays.

        By default, it'll at least return the Chinese New Year holidays chosen
        using the following options:

        * ``include_chinese_new_year_eve``
        * ``include_chinese_new_year`` (on by default)
        * ``include_chinese_second_day``

        If the ``shift_sunday_holidays`` option is on, the rules are the
        following.

        * If the CNY1 falls on MON-FRI, there's not shift.
        * If the CNY1 falls on SAT, the CNY2 is shifted to the Monday after.
        * If the CNY1 falls on SUN, the CNY1 is shifted to the Monday after,
          and CNY2 is shifted to the Tuesday after.
        """
        days = []

        lunar_first_day = ChineseNewYearMixin.lunar(year, 1, 1)
        # Chinese new year's eve
        if self.include_chinese_new_year_eve:
            days.append((
                lunar_first_day - timedelta(days=1),
                self.chinese_new_year_eve_label
            ))
        # Chinese new year (is included by default)
        if self.include_chinese_new_year:
            days.append((lunar_first_day, self.chinese_new_year_label))

        if self.include_chinese_second_day:
            lunar_second_day = lunar_first_day + timedelta(days=1)
            days.append((
                lunar_second_day,
                self.chinese_second_day_label
            ))
        if self.include_chinese_third_day:
            lunar_third_day = lunar_first_day + timedelta(days=2)
            days.append((
                lunar_third_day,
                self.chinese_third_day_label
            ))

        if self.shift_sunday_holidays:
            if lunar_first_day.weekday() == SUN:
                if self.shift_start_cny_sunday:
                    days.append(
                        (lunar_first_day - timedelta(days=1),
                         "Chinese Lunar New Year shift"),
                    )
                else:
                    if self.include_chinese_third_day:
                        shift_day = lunar_third_day
                    else:
                        shift_day = lunar_second_day
                    days.append(
                        (shift_day + timedelta(days=1),
                         "Chinese Lunar New Year shift"),
                    )
            if (lunar_second_day.weekday() == SUN
                    and self.include_chinese_third_day):
                days.append(
                    (lunar_third_day + timedelta(days=1),
                     "Chinese Lunar New Year shift"),
                )
        return days

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.extend(self.get_chinese_new_year(year))
        return days

    def get_shifted_holidays(self, dates):
        """
        Taking a list of existing holidays, yield a list of 'shifted' days if
        the holiday falls on SUN.
        """
        for holiday, label in dates:
            if holiday.weekday() == SUN:
                yield (
                    holiday + timedelta(days=1),
                    f'{label} shift'
                )

    def get_calendar_holidays(self, year):
        """
        Take into account the eventual shift to the next MON if any holiday
        falls on SUN.
        """
        # Unshifted days are here:
        days = super().get_calendar_holidays(year)
        if self.shift_sunday_holidays:
            days_to_inspect = copy(days)
            for day_shifted in self.get_shifted_holidays(days_to_inspect):
                days.append(day_shifted)
        return days


class CalverterMixin:
    conversion_method = None
    ISLAMIC_HOLIDAYS = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.conversion_method is None:
            raise NotImplementedError

    def converted(self, year):
        current = date(year, 1, 1)
        delta = timedelta(days=1)
        days = []
        while current.year == year:
            days.append(
                self.conversion_method.from_gregorian(
                    current.year,
                    current.month,
                    current.day
                )
            )
            current += delta
        return days

    def calverted_years(self, year):
        converted = self.converted(year)
        return sorted({y for y, m, d in converted})

    def get_islamic_holidays(self):
        return self.ISLAMIC_HOLIDAYS

    def get_delta_islamic_holidays(self, year):
        """
        Return the delta to add/substract according to the year or customs.

        By default, to return None or timedelta(days=0)
        """
        return None

    def get_variable_days(self, year):
        warnings.warn('Please take note that, due to arbitrary decisions, '
                      'this Islamic calendar computation may be wrong.')
        days = super().get_variable_days(year)
        years = self.calverted_years(year)
        for month, day, label in self.get_islamic_holidays():
            for y in years:
                g_date = self.conversion_method.to_gregorian(y, month, day)
                holiday = date(*g_date)

                # Only add a delta if necessary
                delta = self.get_delta_islamic_holidays(year)
                if delta:
                    holiday += delta

                if holiday.year == year:
                    days.append((holiday, label))
        return days


class IslamicMixin(CalverterMixin):

    WEEKEND_DAYS = (FRI, SAT)

    conversion_method = convertdate.islamic
    include_prophet_birthday = False
    include_day_after_prophet_birthday = False
    include_start_ramadan = False
    include_eid_al_fitr = False
    length_eid_al_fitr = 1
    eid_al_fitr_label = "Eid al-Fitr"
    include_eid_al_adha = False
    eid_al_adha_label = "Eid al-Adha"
    length_eid_al_adha = 1
    include_day_of_sacrifice = False
    day_of_sacrifice_label = "Eid al-Adha"
    include_islamic_new_year = False
    include_laylat_al_qadr = False
    include_nuzul_al_quran = False

    def get_islamic_holidays(self):
        """Return a list of Islamic (month, day, label) for islamic holidays.
        Please take note that these dates must be expressed using the Islamic
        Calendar"""
        days = list(super().get_islamic_holidays())

        if self.include_islamic_new_year:
            days.append((1, 1, "Islamic New Year"))
        if self.include_prophet_birthday:
            days.append((3, 12, "Prophet's Birthday"))
        if self.include_day_after_prophet_birthday:
            days.append((3, 13, "Day after Prophet's Birthday"))
        if self.include_start_ramadan:
            days.append((9, 1, "Start of ramadan"))
        if self.include_nuzul_al_quran:
            days.append((9, 17, "Nuzul Al-Qur'an"))
        if self.include_eid_al_fitr:
            for x in range(self.length_eid_al_fitr):
                days.append((10, x + 1, self.eid_al_fitr_label))
        if self.include_eid_al_adha:
            for x in range(self.length_eid_al_adha):
                days.append((12, x + 10, self.eid_al_adha_label))
        if self.include_day_of_sacrifice:
            days.append((12, 10, self.day_of_sacrifice_label))
        if self.include_laylat_al_qadr:
            warnings.warn("The Islamic holiday named Laylat al-Qadr is decided"
                          " by the religious authorities. It is not possible"
                          " to compute it. You'll have to add it manually.")
        return tuple(days)


class CoreCalendar:

    FIXED_HOLIDAYS = ()
    WEEKEND_DAYS = ()

    def __init__(self):
        self._holidays = {}

    @classproperty
    def name(cls):
        class_name = cls.__name__
        if cls.__doc__:
            doc = cls.__doc__.split('\n')
            doc = map(lambda s: s.strip(), doc)
            return next(s for s in doc if s)
        return class_name

    def get_fixed_holidays(self, year):
        """Return the fixed days according to the FIXED_HOLIDAYS class property
        """
        days = []
        for month, day, label in self.FIXED_HOLIDAYS:
            days.append((date(year, month, day), label))
        return days

    def get_variable_days(self, year):
        return []

    def get_calendar_holidays(self, year):
        """Get calendar holidays.
        If you want to override this, please make sure that it **must** return
        a list of tuples (date, holiday_name)."""
        return self.get_fixed_holidays(year) + self.get_variable_days(year)

    def holidays(self, year=None):
        """Computes holidays (non-working days) for a given year.
        Return a 2-item tuple, composed of the date and a label."""
        if not year:
            year = date.today().year

        if year in self._holidays:
            return self._holidays[year]

        # Here we process the holiday specific calendar
        temp_calendar = tuple(self.get_calendar_holidays(year))

        # it is sorted
        self._holidays[year] = sorted(temp_calendar)
        return self._holidays[year]

    def get_holiday_label(self, day):
        """Return the label of the holiday, if the date is a holiday"""
        day = cleaned_date(day)
        return {day: label for day, label in self.holidays(day.year)}.get(day)

    def holidays_set(self, year=None):
        "Return a quick date index (set)"
        return {day for day, label in self.holidays(year)}

    def get_weekend_days(self):
        """Return a list (or a tuple) of weekdays that are *not* working days.

        e.g: return (SAT, SUN,)

        """
        if self.WEEKEND_DAYS:
            return self.WEEKEND_DAYS
        else:
            raise NotImplementedError("Your Calendar class must provide "
                                      "WEEKEND_DAYS or implement the "
                                      "`get_weekend_days` method")

    def is_working_day(self, day,
                       extra_working_days=None, extra_holidays=None):
        """Return True if it's a working day.
        In addition to the regular holidays, you can add exceptions.

        By providing ``extra_working_days``, you'll state that these dates
        **are** working days.

        By providing ``extra_holidays``, you'll state that these dates **are**
        holidays, even if not in the regular calendar holidays (or weekends).

        Please note that the ``extra_working_days`` list has priority over the
        ``extra_holidays`` list.

        """
        day = cleaned_date(day)
        if extra_working_days:
            extra_working_days = tuple(map(cleaned_date, extra_working_days))
        if extra_holidays:
            extra_holidays = tuple(map(cleaned_date, extra_holidays))

        # Extra lists exceptions
        if extra_working_days and day in extra_working_days:
            return True

        # Regular rules
        if day.weekday() in self.get_weekend_days():
            return False

        return not self.is_holiday(day, extra_holidays=extra_holidays)

    def is_holiday(self, day, extra_holidays=None):
        """Return True if it's an holiday.
        In addition to the regular holidays, you can add exceptions.

        By providing ``extra_holidays``, you'll state that these dates **are**
        holidays, even if not in the regular calendar holidays (or weekends).

        """
        day = cleaned_date(day)

        if extra_holidays:
            extra_holidays = tuple(map(cleaned_date, extra_holidays))

        if extra_holidays and day in extra_holidays:
            return True

        return day in self.holidays_set(day.year)

    def add_working_days(self, day, delta,
                         extra_working_days=None, extra_holidays=None,
                         keep_datetime=False):
        """Add `delta` working days to the date.

        You can provide either a date or a datetime to this function that will
        output a ``date`` result. You can alter this behaviour using the
        ``keep_datetime`` option set to ``True``.

        the ``delta`` parameter might be positive or negative. If it's
        negative, you may want to use the ``sub_working_days()`` method with
        a positive ``delta`` argument.

        By providing ``extra_working_days``, you'll state that these dates
        **are** working days.

        By providing ``extra_holidays``, you'll state that these dates **are**
        holidays, even if not in the regular calendar holidays (or weekends).

        Please note that the ``extra_working_days`` list has priority over the
        ``extra_holidays`` list.
        """
        day = cleaned_date(day, keep_datetime)

        if extra_working_days:
            extra_working_days = tuple(map(cleaned_date, extra_working_days))

        if extra_holidays:
            extra_holidays = tuple(map(cleaned_date, extra_holidays))

        days = 0
        temp_day = day
        day_added = 1 if delta >= 0 else -1
        delta = abs(delta)
        while days < delta:
            temp_day = temp_day + timedelta(days=day_added)
            if self.is_working_day(temp_day,
                                   extra_working_days=extra_working_days,
                                   extra_holidays=extra_holidays):
                days += 1
        return temp_day

    def sub_working_days(self, day, delta,
                         extra_working_days=None, extra_holidays=None,
                         keep_datetime=False):
        """
        Substract `delta` working days to the date.

        This method is a shortcut / helper. Users may want to use either::

            cal.add_working_days(my_date, -7)
            cal.sub_working_days(my_date, 7)

        The other parameters are to be used exactly as in the
        ``add_working_days`` method.

        A negative ``delta`` argument will be converted into its absolute
        value. Hence, the two following calls are equivalent::

            cal.sub_working_days(my_date, -7)
            cal.sub_working_days(my_date, 7)

        As in ``add_working_days()`` you can set the parameter
        ``keep_datetime`` to ``True`` to make sure that if your ``day``
        argument is a ``datetime``, the returned date will also be a
        ``datetime`` object.

        """
        delta = abs(delta)
        return self.add_working_days(
            day, -delta,
            extra_working_days, extra_holidays, keep_datetime=keep_datetime)

    def find_following_working_day(self, day):
        """Looks for the following working day, if not already a working day.

        **WARNING**: this function doesn't take into account the calendar
        holidays, only the days of the week and the weekend days parameters.
        """
        day = cleaned_date(day)

        while day.weekday() in self.get_weekend_days():
            day = day + timedelta(days=1)
        return day

    @staticmethod
    def get_nth_weekday_in_month(year, month, weekday, n=1, start=None):
        """Get the nth weekday in a given month. e.g:

        >>> # the 1st monday in Jan 2013
        >>> Calendar.get_nth_weekday_in_month(2013, 1, MON)
        datetime.date(2013, 1, 7)
        >>> # The 2nd monday in Jan 2013
        >>> Calendar.get_nth_weekday_in_month(2013, 1, MON, 2)
        datetime.date(2013, 1, 14)
        """
        # If start is `None` or Falsy, no need to check and clean
        if start:
            start = cleaned_date(start)

        day = date(year, month, 1)
        if start:
            day = start
        counter = 0
        while True:
            if day.month != month:
                # Don't forget to break if "n" is too big
                return None
            if day.weekday() == weekday:
                counter += 1
            if counter == n:
                break
            day = day + timedelta(days=1)
        return day

    @staticmethod
    def get_last_weekday_in_month(year, month, weekday):
        """Get the last weekday in a given month. e.g:

        >>> # the last monday in Jan 2013
        >>> Calendar.get_last_weekday_in_month(2013, 1, MON)
        datetime.date(2013, 1, 28)
        """
        day = date(year, month, monthrange(year, month)[1])
        while True:
            if day.weekday() == weekday:
                break
            day = day - timedelta(days=1)
        return day

    @staticmethod
    def get_iso_week_date(year, week_nb, weekday=ISO_MON):
        """
        Return the date of the weekday of the week number (ISO definition).

        **Warning:** in the ISO definition, the weeks start on MON, not SUN.

        By default, if you don't provide the ``weekday`` argument, it'll return
        the date of the MON of this week number.

        Example:

            >>> Calendar.get_iso_week_date(2021, 44)
            datetime.date(2021, 11, 1)

        For your convenience, the ISO weekdays are available via the
        ``workalendar.core`` module, like this:

            from workalendar.core import ISO_MON, ISO_TUE  # etc.

        i.e.: if you need to get the FRI of the week 44 of the year 2020,
        you'll have to use:

            from workalendar.core import ISO_FRI
            Calendar.get_iso_week_date(2020, 44, ISO_FRI)

        """
        if sys.version_info >= (3, 8, 0):
            # use the stock Python 3.8 function
            return date.fromisocalendar(year, week_nb, weekday)

        # else, use the backport
        # Adapted from https://stackoverflow.com/a/59200842
        jan_1st = date(year, 1, 1)
        _, jan_1st_week, jan_1st_weekday = jan_1st.isocalendar()
        base = 1 if jan_1st_week == 1 else 8
        delta = base - jan_1st_weekday + 7 * (week_nb - 1) + (weekday - 1)
        start = jan_1st + timedelta(days=delta)
        return start

    @staticmethod
    def get_first_weekday_after(day, weekday):
        """Get the first weekday after a given day. If the day is the same
        weekday, the same day will be returned.

        >>> # the first monday after Apr 1 2015
        >>> Calendar.get_first_weekday_after(date(2015, 4, 1), MON)
        datetime.date(2015, 4, 6)

        >>> # the first tuesday after Apr 14 2015
        >>> Calendar.get_first_weekday_after(date(2015, 4, 14), TUE)
        datetime.date(2015, 4, 14)
        """
        day_delta = (weekday - day.weekday()) % 7
        day = day + timedelta(days=day_delta)
        return day

    def get_working_days_delta(self, start, end, include_start=False,
                               extra_working_days=None, extra_holidays=None):
        """
        Return the number of working day between two given dates.
        The order of the dates provided doesn't matter.

        In the following example, there are 5 days, because of the week-end:

        >>> cal = WesternCalendar()  # does not include easter monday
        >>> day1 = date(2018, 3, 29)
        >>> day2 = date(2018, 4, 5)
        >>> cal.get_working_days_delta(day1, day2)
        5

        In France, April 1st 2018 is a holiday because it's Easter monday:

        >>> cal = France()
        >>> cal.get_working_days_delta(day1, day2)
        4

        This method should even work if your ``start`` and ``end`` arguments
        are datetimes.

        By default, if the day after you start is not a working day,
        the count will start at 0. If include_start is set to true,
        this day will be taken into account.

        Example:

        >>> from dateutil.parser import parse
        >>> cal = France()
        >>> day1 = parse('09/05/2018 00:01', dayfirst=True)
        >>> day2 = parse('10/05/2018 19:01', dayfirst=True) # holiday in france
        >>> cal.get_working_days_delta(day1, day2)
        0

        >>> cal.get_working_days_delta(day1, day2, include_start=True)
        1

        As in many other methods, you can use the ``extra_holidays`` and
        ``extra_working_days`` to exclude
        """
        start = cleaned_date(start)
        end = cleaned_date(end)

        if start == end:
            return 0

        if start > end:
            start, end = end, start

        # Starting count here
        is_working_day = self.is_working_day(
            start,
            extra_working_days=extra_working_days,
            extra_holidays=extra_holidays
        )
        count = 1 if include_start and is_working_day else 0
        while start < end:
            start += timedelta(days=1)
            is_working_day = self.is_working_day(
                start,
                extra_working_days=extra_working_days,
                extra_holidays=extra_holidays
            )
            if is_working_day:
                count += 1
        return count

    def _get_ical_period(self, period=None):
        """
        Return a usable period for iCal export

        Default period is [2000, 2030]
        """
        # Default value.
        if not period:
            period = [2000, 2030]

        # Make sure it's a usable iterable
        if type(period) not in (list, tuple):
            raise ICalExportRangeError(
                "Incorrect Range type. Must be list or tuple.")

        # Taking the extremes
        period = [min(period), max(period)]

        # check for internal types
        check_types = map(type, period)
        check_types = map(lambda x: x != int, check_types)
        if any(check_types):
            raise ICalExportRangeError(
                "Incorrect Range boundaries. Must be int.")

        return period

    def _get_ical_target_path(self, target_path):
        """
        Return target path for iCal export.

        Note
        ----
        If `target_path` does not have one of the extensions `.ical`, `.ics`,
        `.ifb`, or `.icalendar`, the extension `.ics` is appended to the path.
        Returns
        -------
        None.
        Examples
        --------
        >>> cal = Austria()
        >>> cal._get_ical_target_path('austria')  # -> austria.ics

        """

        if not target_path:
            raise ICalExportTargetPathError(
                "Incorrect target path. It must not be empty")

        target_path = Path(target_path)

        if target_path.is_dir():
            raise ICalExportTargetPathError(
                "Incorrect target path. It must not be a directory"
            )

        ical_extensions = ['.ical', '.ics', '.ifb', '.icalendar']
        if target_path.suffix not in ical_extensions:
            target_path = target_path.with_name(target_path.name + '.ics')
        return target_path

    def export_to_ical(self, period=[2000, 2030], target_path=None):
        """
        Export the calendar to iCal (RFC 5545) format.

        Parameters
        ----------
        period: [int, int]
            start and end year (inclusive) of calendar
            Default is [2000, 2030]

        target_path: str or pathlib.Path
            the name or path of the exported file. If this argument is missing,
            the function will return the ical content.

        """
        first_year, last_year = self._get_ical_period(period)
        if target_path:
            # Generate filename path before calculate the holidays
            target_path = self._get_ical_target_path(target_path)

        # fetch holidays
        holidays = []
        for year in range(first_year, last_year + 1):
            holidays.extend(self.holidays(year))

        # initialize icalendar
        ics = [
            'BEGIN:VCALENDAR',
            'VERSION:2.0',  # current RFC5545 version
            f'PRODID:-//workalendar//ical {__version__}//EN'
        ]
        dtstamp = f'DTSTAMP;VALUE=DATE-TIME:{datetime.utcnow():%Y%m%dT%H%M%SZ}'

        # add an event for each holiday
        for date_, name in holidays:
            ics.extend([
                'BEGIN:VEVENT',
                f'SUMMARY:{name}',
                f'DTSTART;VALUE=DATE:{date_:%Y%m%d}',
                dtstamp,
                f'UID:{date_}{name}@peopledoc.github.io/workalendar',
                'END:VEVENT',
            ])

        # add footer
        ics.append('END:VCALENDAR\n')  # last line with a trailing \n

        # Transform this list into text lines
        ics = "\n".join(ics)

        if target_path:
            # save iCal file
            with target_path.open('w+') as export_file:
                export_file.write(ics)
            return

        return ics


class Calendar(CoreCalendar):
    """
    The cornerstone of Earth calendars.

    Take care of the New Years Day, which is almost a worldwide holiday.
    """
    include_new_years_day = True
    include_new_years_eve = False
    shift_new_years_day = False
    include_labour_day = False
    labour_day_label = "Labour Day"

    def __init__(self, **kwargs):
        super().__init__()

    def get_fixed_holidays(self, year):
        days = super().get_fixed_holidays(year)
        if self.include_new_years_day:
            days.insert(0, (date(year, 1, 1), "New year"))
        if self.include_new_years_eve:
            days.append((date(year, 12, 31), "New Year's eve"))
        if self.include_labour_day:
            days.append((date(year, 5, 1), self.labour_day_label))
        return days

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        new_year = date(year, 1, 1)
        if self.include_new_years_day and self.shift_new_years_day:
            if new_year.weekday() in self.get_weekend_days():
                days.append((
                    self.find_following_working_day(new_year),
                    "New Year shift"))
        return days


class WesternCalendar(WesternMixin, Calendar):
    """
    A Christian calendar using Western definition for Easter.
    """


class OrthodoxCalendar(OrthodoxMixin, Calendar):
    """
    A Christian calendar using Orthodox definition for Easter.
    """


class ChineseNewYearCalendar(ChineseNewYearMixin, Calendar):
    """
    Chinese Calendar, using Chinese New Year computation.
    """
    # There are regional exceptions to those week-end days, to define locally.
    WEEKEND_DAYS = (SAT, SUN)


class IslamicCalendar(IslamicMixin, Calendar):
    """
    Islamic calendar
    """


class IslamoWesternCalendar(IslamicMixin, WesternMixin, Calendar):
    """
    Mix of Islamic and Western calendars.

    When countries have both Islamic and Western-Christian holidays.
    """
    FIXED_HOLIDAYS = Calendar.FIXED_HOLIDAYS
